/*****************************************************************************
 *                                                                           *
 *        **********                                                         *
 *       ************                                                        *
 *      ***        ***                                                       *
 *      ***   +++   ***                                                      *
 *      ***   + +   ***                                                      *
 *      ***   +                    CHIPCON CC1010 EXAMPLE PROGRAM            *
 *      ***   + +   ***                      CC1010 CHAT                     *
 *      ***   +++   ***                                                      *
 *      ***       ***                                                        *
 *       ***********                                                         *
 *        *********                                                          *
 *                                                                           *
 *****************************************************************************
 * NOTE:                                                                     *
 * This program has exactly the same functionality as the original chat      *
 * example program, but it has been modified to use the Frequency Hopping    *
 * Protocol (FHP) instead of the Simple Packet Protocol (SPP) through the    *
 * use of the Emulator (EMU).                                                *
 * This has been done simply by performing the following substitutions:      *
 *  - sppSetupRF(); --> initializeFHP(TRUE/FALSE); setMode(ACTIVE);          *
 *  - sppSend       --> fhsppSend                                            *
 *  - sppReceive    --> fhsppReceive                                         *
 *  - sppStatus     --> fhsppStatus                                          *
 *  - sppReset      --> fhsppReset                                           *
 * END OF NOTE                                                               *
 *                                                                           *
 * This program demonstrates the use of the SPP library, random number       *
 * generation, timers, the serial port (with use of escape codes), how to    *
 * assign SPP addresses and maintain a client table, pinging (broadcasted to *
 * all clients at regular intervals), and DES encryption.                    *
 *                                                                           *
 * INSTRUCTIONS:                                                             *
 * Use up to 16 evaluation boards, and connect each of them to a terminal    *
 * window via serial port 0 (38400). Create a server and 1-15 clients.       *
 * Please follow the instructions displayed at startup.                      *  
 *                                                                           *
 * HOW IT WORKS:                                                             *
 * The server and the clients share the same code. The SPP address (255 for  *
 * servers, 254 for unregistered clients and 1-15 for clients) is used to    *
 * distinguish between them.                                                 *
 *                                                                           *
 * The chat server maintains and distributes a client table. All clients     *
 * have a copy of this table, which contain nick names, addresses, keys,     *
 * etc. The server also transmits a ping message at regular intervals (about * 
 * every second). If no ping or other server messages have been received     *
 * during the last 20 seconds, the client will automatically disconnect      *
 * (passively). It will also be disconnected (actively) if messages from the *
 * server don't get through. To avoid jamming of server messages, the server *
 * has 50 TX attempts available, while the clients only have two.            *
 *                                                                           *
 * The text input is handled by the UART0 ISR. When ENTER is pressed, a flag *
 * is set to indicate that there is message ready for transmission.          *
 *                                                                           *
 * The main program loop will receive and process incoming messages. When    *
 * the TX flag is on, it will transmit, both broadcasted and private         *
 * messages. Key exchange is performed if necessary.                         *
 *                                                                           *
 * The printf/putchar functions are used to print text to the chat window.   *
 * This requires switching between interrupt-driven and polled I/O.          *
 *                                                                           *
 * LEDs:                                                                     *
 *     RED:    Retransmission                                                *
 *     YELLOW: Transmission                                                  *
 *     GREEN:  Processing received message						             *
 *     BLUE:   Toggled for each received ping                     		     *
 *                                                                           *
 * NOTE: This program requires HyperTerminal 5.0 or later (bug fix).         *
 *****************************************************************************
 * Author:              JOL                                                  *
 *****************************************************************************
 * Revision history:                                                         *
 * 1.0  2002/08/13      First public release                                 *
 *                                                                           *
 * $Log: chatMain.c,v $
 * Revision 1.1  2003/07/29 11:16:30  tos
 * Initial version in CVS.
 *
 *
 *                                                                           *
 ****************************************************************************/

#include "chat.h"

// Data buffers
byte xdata pMessageTXBuffer[BUFFER_LENGTH]	 			_at_ 0x0000; // Transmitted messages
byte xdata pAdminTXBuffer[BUFFER_LENGTH] 				_at_ 0x0100; // Transmitted administration messages
byte xdata pRXBuffer[BUFFER_LENGTH]		 				_at_ 0x0200; // Received messages 
byte xdata pKeyBuffer[CI_KEY_LENGTH * MAX_CLIENTS]	 	_at_ 0x0500; // DES keys

// The current input position (equals the length of the typed string)
byte xdata inPos;

// The ping countdown variable (clients only!)
word xdata noPingTicks;

// Disables message decryption (for demonstration purposes)
byte xdata noDecryption;

// The client info table
CLIENT_INFO xdata CI[MAX_CLIENTS];
CLIENT_INFO xdata *pMyCI;
byte xdata clientCount;

// SPP data
SPP_SETTINGS xdata sppSettings;
SPP_RX_INFO xdata RXI;
SPP_TX_INFO xdata TXI;

// Transmission request flag (transmit asap)
byte xdata txRequest;
#define TX_REQUEST_ON  1
#define TX_REQUEST_OFF 0

// Private (encrypted) messages
byte xdata privateAddress;	
byte xdata isPrivateMode;

//----------------------------------------------------------------------------
//	CLIENT MAIN PROGRAM: setup -> login -> main loop
//----------------------------------------------------------------------------
void main (void) {

	// Local variables
	byte temp, n;
	CLIENT_INFO xdata *pCI;

    // Startup macros for speed and low power consumption
    MEM_NO_WAIT_STATES();
    FLASH_SET_POWER_MODE(FLASH_STANDBY_BETWEEN_READS);

	// Initialize peripherals
	WDT_ENABLE(FALSE);
	RLED_OE(TRUE);
	YLED_OE(TRUE);
	GLED_OE(TRUE);
	BLED_OE(TRUE);

	// Initialize the client info table
	initClientInfo();

	// Setup serial port flow control
    PORTDIRBIT(1, 5, POUT);
	UART0_FLOW_CONTROL = READY;

    // Setup UART0 for polled I/O 
    UART0_SETUP(38400, CLK_FREQ, UART_NO_PARITY | UART_RX_TX | UART_POLLED);

	// Initialize the login screen
	VT100_INIT_LOGIN_SCREEN();
	printf("********************************************************************************");
	printf("                                   CC1010 CHAT                                  ");
	printf("********************************************************************************");
	printf("INSTRUCTIONS:");
	printf("\n  - Broadcasted messages are not encrypted");
	printf("\n  - To send a broadcasted message, type: [message][ENTER]");
	printf("\n  - The maximum length of a message is one line");
	printf("\n  - The DELETE key and the arrow keys are not available"); 
	printf("\n");
	printf("\nSECURITY:");
	printf("\n  - Private messages are encrypted with HW single DES (Dark yellow background)");
	printf("\n  - To send a private message, type [client index (HEX) + '<'][message][RETURN]");
	printf("\n  - To enable/disable message decryption, type ['<']\n");
	printf("********************************************************************************");
	printf("\n");

	// Ask the user for a nick name
	printf("Enter your nick name (%d chars max): ", (int) (NAME_LENGTH - 1)); // Remember the null terminator
	memset(pMyCI->name, '\0', sizeof(pMyCI->name));	
	gets(pMyCI->name, NAME_LENGTH);

	// Ask the user to choose between server and client	
	printf("\nPress '1' to become a client or '2' to become a server: ");  
	while(TRUE) {
		if ((temp = getchar()) == '2') {
	
			// Defualt local settings for servers
			sppSettings.myAddress = pMyCI->address = SERVER_ADDRESS;
			sppSettings.rxTimeout = LONG_RX_TIMEOUT;
			sppSettings.txAckTimeout = 2;
			sppSettings.txAttempts = 50;
			sppSettings.txPreambleByteCount = 7;

            // RF setup
//            sppSetupRF(&RF_SETTINGS, &RF_CALDATA, TRUE);
			initializeFHP(TRUE);
			setMode(ACTIVE);

            // Initialize the SPP timer
            sppStartTimer(CC1010EB_CLKFREQ);
            SPP_INIT_TIMEOUTS();

			printf("\nSERVER: Creating chat Room...");
//			halWait(255, CLK_FREQ);
			break;

		} else if (temp == '1') {
	
			// Defualt local settings for clients (the address is temporary)
			sppSettings.myAddress = pMyCI->address = NEW_CLIENT_ADDRESS;
			sppSettings.rxTimeout = LONG_RX_TIMEOUT;
			sppSettings.txAckTimeout = 2;
			sppSettings.txAttempts = 2;
			sppSettings.txPreambleByteCount = 7;

            // RF setup
//            fhsppSetupRF(&RF_SETTINGS, &RF_CALDATA, TRUE);
			initializeFHP(FALSE);
			setMode(ACTIVE);

            // Initialize the SPP timer
            sppStartTimer(CC1010EB_CLKFREQ);
            SPP_INIT_TIMEOUTS();

			printf("\nCLIENT: Joining chat room...");
//			halWait(255,CLK_FREQ);

			// Join request
			while ((temp = sendJoinRequest()) != JOIN_ACCEPTED) {
				if (temp == JOIN_DENIED) {
					printf("\nInvalid nick name / The server is full!");
					while (TRUE);
				} else {
					printf("\nJoin error (RF related)");
					halWait(255, CLK_FREQ);
					printf(" -> retrying now...");
					halWait(255, CLK_FREQ);
				}
			}
			sppSettings.myAddress = pMyCI->address;

			// Initialize the ping timeout (we'll reset the countdown each 
			// time we receive a ping or some other message from the server
			noPingTicks = NO_PING_TIMEOUT;
			sppSetTimerCB(SPP_CUSTOM_0_TIMER, disconnected, &noPingTicks);

			break;
		}
	}

	// Initialize the chat screen
	VT100_INIT_CHAT_SCREEN();

	// Setup UART0 for interrupt driven I/O 
    UART0_SETUP(38400, CLK_FREQ, UART_NO_PARITY | UART_RX_TX | UART_ISR);
//	printf("Interrupt UART /1\n");
	INT_SETFLAG(INUM_UART0_TX, INT_SET);

//	printf("Interrupt UART /2\n");

	// Initialize some chat variables
	inPos = 0;
	TXI.status = SPP_TX_FINISHED;
	txRequest = TX_REQUEST_OFF;
	noDecryption = FALSE;
	isPrivateMode = FALSE;

//	printf("printClientList...\n");
	printClientList();
//	printf("printClientList done\n");

	// Loop forever
	while (TRUE) {

		// Turn on reception
		RXI.maxDataLen = BUFFER_LENGTH;
		RXI.pDataBuffer = pRXBuffer;

		fhsppReceive(&RXI);

		// Wait for RF RX to complete
		do {
			if (txRequest && (TXI.status == SPP_TX_FINISHED) && (RXI.status == SPP_RX_WAITING)) {
				fhsppReset();
			}
		} while (fhsppStatus() != SPP_IDLE_MODE); 

		// Process incoming messages, if any
		if (RXI.status == SPP_RX_FINISHED) {
			YLED = LED_ON;
			processMessage();
			YLED = LED_OFF;
		}

		// Server transmits ping
		if (!txRequest && (pMyCI->address == SERVER_ADDRESS)) {
			YLED = LED_ON;

			// Prepare the transmission settings
			TXI.destination = SPP_BROADCAST;
			TXI.dataLen = 2;
			TXI.pDataBuffer = pAdminTXBuffer;
			TXI.flags = MSG_PING;

			// Reset the 16-bit array
			pAdminTXBuffer[0] = 0x00;
			pAdminTXBuffer[1] = 0x00;

			// pAdminTXBuffer[0]  pAdminTXBuffer[1]
			// MSB(15) ... LSB(8) MSB(7) ... LSB(0)
			for (n = 1; n < 16; n++) {
				if (CI[n].address == INVALID_ADDRESS) {
					// Do nothing
				} else if (CI[n].address < 8) {
					pAdminTXBuffer[1] |= (0x01 << CI[n].address);
				} else { // CI[n].address >= 8
					pAdminTXBuffer[0] |= (0x01 << (CI[n].address - 8));
				}
			}

			// Send it
			fhsppSend(&TXI);
			while (fhsppStatus() != SPP_IDLE_MODE);
			YLED = LED_OFF;
		}

		// Transmit outgoing messages 
		if (txRequest) {
			YLED = LED_ON;

			// PRIVATE MESSAGE
			if (isPrivateMode) {

				// Make sure that the DES key is valid
				pCI = getClientInfo(privateAddress);
				if (!pCI->isDESKeyValid) {
					
					// Discard the message if the key exchange fails.
//					debu(sprintf(mess,"main %X",(word)privateAddress));
					if (!diffieHellmanUp(privateAddress)) {
						printStatus(KEY_EXCHANGE_FAILED, privateAddress);
						clearInputSection();
						isPrivateMode = FALSE;
						inPos = 0;
						txRequest = TX_REQUEST_OFF;
						UART0_FLOW_CONTROL = READY;
					} else {
						printStatus(KEY_EXCHANGE_OK, privateAddress);
					}						
				}

				if (pCI->isDESKeyValid && getClientInfo(privateAddress)) {

					// Prepare the transmission settings
					TXI.destination = privateAddress;
					TXI.flags = (pCI->txFlags & SPP_SEQUENCE_BIT) | SPP_ACK_REQ | SPP_ENCRYPTED_DATA | MSG_CHAT_DOWN;
					TXI.dataLen = inPos;
					TXI.pDataBuffer = pMessageTXBuffer;

					// Encrypt the text with single DES
					halDES(DES_SINGLE_DES | DES_ENCRYPT | DES_OFB_MODE, TXI.pDataBuffer, pCI->pDESKey, TXI.dataLen);

					// Send it
					fhsppSend(&TXI);
					while (fhsppStatus() != SPP_IDLE_MODE);

					// If OK, print the message, clear the input line, and get ready for more input
					if (TXI.status == SPP_TX_FINISHED) {
						pCI->txFlags = TXI.flags;
						sppSettings.rxTimeout = LONG_RX_TIMEOUT;

						// We'll have to decrypt the message to be able to print it
						halDES(DES_SINGLE_DES | DES_DECRYPT | DES_OFB_MODE, TXI.pDataBuffer, pCI->pDESKey, TXI.dataLen);
						printMessage(pMyCI->name, isPrivateMode, pMessageTXBuffer, TXI.dataLen);
						clearInputSection();
						isPrivateMode = FALSE;
						inPos = 0;
						txRequest = TX_REQUEST_OFF;
						UART0_FLOW_CONTROL = READY;
						RLED = LED_OFF;
					} else {
						sppSettings.rxTimeout = RETRY_TX_TIMEOUT;
						RLED = LED_ON;
					}
				}

			// SERVER BROADCAST
			} else if (pMyCI->address == SERVER_ADDRESS) {

				// Send the message to all clients
				broadcastMessage(SERVER_ADDRESS, pMessageTXBuffer, inPos);

				// Print the message, clear the input line, and get ready for more input
				printMessage(pMyCI->name, isPrivateMode, pMessageTXBuffer, inPos);
				clearInputSection();
				inPos = 0;
				txRequest = TX_REQUEST_OFF;
				UART0_FLOW_CONTROL = READY;

			// CLIENT BROADCAST (VIA SERVER)
			} else {
				// Prepare the transmission settings
				TXI.destination = SERVER_ADDRESS;
				TXI.flags = (getClientInfo(SERVER_ADDRESS)->txFlags & SPP_SEQUENCE_BIT) | SPP_ACK_REQ | MSG_CHAT_UP;
				TXI.dataLen = inPos;
				TXI.pDataBuffer = pMessageTXBuffer;

				// Send the message to the server
				fhsppSend(&TXI);
				while (fhsppStatus() != SPP_IDLE_MODE);

				// If OK, print the message, clear the input line, and get ready for more input
				if (TXI.status == SPP_TX_FINISHED) {
					getClientInfo(SERVER_ADDRESS)->txFlags = TXI.flags;
					sppSettings.rxTimeout = LONG_RX_TIMEOUT;
					printMessage(pMyCI->name, isPrivateMode, pMessageTXBuffer, inPos);
					clearInputSection();
					inPos = 0;
					txRequest = TX_REQUEST_OFF;
					UART0_FLOW_CONTROL = READY;
					RLED = LED_OFF;
				} else {
					sppSettings.rxTimeout = RETRY_TX_TIMEOUT;
					RLED = LED_ON;
				}
			}
			YLED = LED_OFF;
		}
	}
} // main

byte xdata _power[] = "POWER MODE";
byte xdata _illegal[] = "MASTER IS NOT ALLOWED TO SHUT DOWN!";
byte xdata _off[] = "OFF";
byte xdata _passive[] = "PASSIVE";
byte xdata _active[] = "ACTIVE";


//----------------------------------------------------------------------------
//  UART0 0 interrupt service routine: Message input
//----------------------------------------------------------------------------
void UART0_ISR () interrupt INUM_UART0 {
	char typedChar;

	// Ignore TX interrupts
	if (TI_0) {
		INT_SETFLAG(INUM_UART0_TX, INT_CLR);
		return;
	}	

	if (RI_0) {
		INT_SETFLAG(INUM_UART0_RX, INT_CLR);

		// ENTER transmits the message (if there is any)
		if ((typedChar = UART0_RECEIVE()) == TERMINAL_KEY_ENTER) {
			if (inPos > 0) {
				// Send the message
				UART0_FLOW_CONTROL = STOP;
				txRequest = TX_REQUEST_ON;
			}

		// Max input length exceeded, or ENTER has already been hit
		} else if ((inPos >= MAX_INPUT_LENGTH) || txRequest) {
			// Do nothing :)

		// BACKSPACE deletes the last character (go back, print space, go back again)
		} else if (typedChar == TERMINAL_KEY_BACKSPACE) {
			if (inPos > 0) {
				INT_SETFLAG(INUM_UART0_TX, INT_SET);
// Results in warning message because of multiple calls, but should not be a problem
				VT100_GO_TO_POS(20, inPos);
				putchar(' ');
// Results in warning message because of multiple calls, but should not be a problem
				VT100_GO_TO_POS(20, inPos);
				inPos--;
			}

		// NORMAL (printable) characters 
		} else if (isprint(typedChar)) {
			INT_SETFLAG(INUM_UART0_TX, INT_SET);

			// Put the character into the TX buffer
			pMessageTXBuffer[inPos++] = typedChar;

			// Private mode check (Input begins with "##<")
			if (inPos == 3) {
				if (pMessageTXBuffer[2] == '<') {
					if (isxdigit(pMessageTXBuffer[0]) && isxdigit(pMessageTXBuffer[1])) {
						privateAddress = strtoul(pMessageTXBuffer, NULL, 16);
//						debu(sprintf(mess,"priv %X %u",(word)privateAddress,(word)privateAddress));
						if ((getClientInfo(privateAddress) != NULL) && (privateAddress != pMyCI->address)) {
							isPrivateMode = TRUE;
						} else {
							isPrivateMode = FALSE;
						}
					} else {
						isPrivateMode = FALSE;
					}
				} else {
					isPrivateMode = FALSE;
				}
			}
			// Make it possible to turn off/on the message decryption
			if ((inPos == 1) && (pMessageTXBuffer[0] == '<')) {
				noDecryption = ~noDecryption;
				inPos = 0;
// Results in warning message because of multiple calls, but should not be a problem
				printStatus(DECRYPTION_MODE, NULL);
				clearInputSection();
			} else {
				// Local echo
// Results in warning message because of multiple calls, but should not be a problem
				VT100_GO_TO_POS(20, inPos);
				putchar(typedChar);
			}

// calling setMode from here results in warning because it is also called from main, should not be problematic
			if ((inPos == 2) && (pMessageTXBuffer[0] == '#')) {
				if (pMyCI->address == SERVER_ADDRESS) {
					printMessage(_power,FALSE,_illegal,strlen(_illegal));
				} else {
					switch(pMessageTXBuffer[1]) {
						case '0': setMode(OFF); printMessage(_power,FALSE,_off,strlen(_off)); break;
						case '1': setMode(PASSIVE); printMessage(_power,FALSE,_passive,strlen(_passive)); break;
						case '2': setMode(ACTIVE); printMessage(_power,FALSE,_active,strlen(_active)); break;
					}
				}
				inPos = 0;
				clearInputSection();
			}
		}

		// Turn off the UART TX interrupt flag (it will be high because of the putchar() calls)
		INT_SETFLAG(INUM_UART0_TX, INT_CLR);
	}
} // UART0_ISR




//----------------------------------------------------------------------------
//  void disconnected (void)
//
//  Description:
//      This function is used by clients when the communication with the 
//		server has broken down.
//----------------------------------------------------------------------------
void disconnected (void) {
	printStatus(DISCONNECTED, NULL);
	while (TRUE);
} // disconnected

